<?php
namespace Incoviba\Controller\API;

use Incoviba\Common\Implement\Exception\EmptyRedis;
use Incoviba\Common\Implement\Exception\EmptyResult;
use Incoviba\Controller\withRedis;
use Incoviba\Exception\ServiceAction\Read;
use Incoviba\Model;
use Incoviba\Repository;
use Incoviba\Service;
use Psr\Http\Message\ResponseInterface;
use Psr\Http\Message\ServerRequestInterface;

class Proyectos
{
    use withJson, withRedis;

    public function list(ServerRequestInterface $request, ResponseInterface $response, Service\Redis $redisService,
                         Repository\Proyecto $proyectoRepository): ResponseInterface
    {
        $output = ['total' => 0, 'proyectos' => []];
        $redisKey = 'proyectos:activos';
        try {
            $output['proyectos'] = $this->fetchRedis($redisService, $redisKey);
            $output['total'] = count($output['proyectos']);
        } catch (EmptyRedis) {
            try {
                $output['proyectos'] = $proyectoRepository->fetchAllActive();
                $output['total'] = count($output['proyectos']);
                $this->saveRedis($redisService, $redisKey, $output['proyectos']);
            } catch (EmptyResult) {}
        }
        return $this->withJson($response, $output);
    }
    public function escriturando(ServerRequestInterface $request, ResponseInterface $response,
                                 Service\Redis $redisService, Repository\Proyecto $proyectoRepository): ResponseInterface
    {
        $output = ['total' => 0, 'proyectos' => []];
        $redisKey = 'proyectos:escriturando';
        try {
            $output['proyectos'] = $this->fetchRedis($redisService, $redisKey);
            $output['total'] = count($output['proyectos']);
        } catch (EmptyRedis) {
            try {
                $output['proyectos'] = $proyectoRepository->fetchAllEscriturando();
                $output['total'] = count($output['proyectos']);
                $this->saveRedis($redisService, $redisKey, $output['proyectos']);
            } catch (EmptyResult) {}
        }
        return $this->withJson($response, $output);
    }
    public function unidades(ServerRequestInterface $request, ResponseInterface $response,
                             Repository\Venta\Unidad $unidadRepository, Service\Redis $redisService,
                             int $proyecto_id): ResponseInterface
    {
        $output = ['proyecto_id' => $proyecto_id, 'unidades' => [], 'total' => 0];
        $redisKey = "unidades:proyecto:{$proyecto_id}";
        try {
            $output = $this->fetchRedis($redisService, $redisKey);
        } catch (EmptyRedis) {
            try {
                $unidades = $unidadRepository->fetchByProyecto($proyecto_id);
                $tipos = [];
                foreach ($unidades as $unidad) {
                    if (!isset($tipos[$unidad->proyectoTipoUnidad->tipoUnidad->descripcion])) {
                        $tipos[$unidad->proyectoTipoUnidad->tipoUnidad->descripcion] = [];
                    }
                    $tipos[$unidad->proyectoTipoUnidad->tipoUnidad->descripcion] []= $unidad;
                }
                foreach ($tipos as &$subtipo) {
                    usort($subtipo, function(Model\Venta\Unidad $a, Model\Venta\Unidad $b) {
                        return strcmp(str_pad($a->descripcion, 4, '0', STR_PAD_LEFT), str_pad($b->descripcion, 4, '0', STR_PAD_LEFT));
                    });
                }
                $output['unidades'] = $tipos;
                $output['total'] = count($unidades);
                $this->saveRedis($redisService, $redisKey, $output);
            } catch (EmptyResult) {}
        }
        return $this->withJson($response, $output);
    }
    public function disponibles(ServerRequestInterface $request, ResponseInterface $response,
                                Service\Redis $redisService, Repository\Venta\Unidad $unidadRepository,
                                int $proyecto_id): ResponseInterface
    {
        $output = [
            'proyecto_id' => $proyecto_id,
            'unidades' => []
        ];
        $redisKey = "unidades:disponibles:proyecto:{$proyecto_id}";
        try {
            $output['unidades'] = $this->fetchRedis($redisService, $redisKey);
        } catch (EmptyRedis) {
            try {
                $output['unidades'] = $unidadRepository->fetchDisponiblesByProyecto($proyecto_id);
                $this->saveRedis($redisService, $redisKey, $output['unidades']);
            } catch (EmptyResult) {}
        }
        return $this->withJson($response, $output);
    }
    public function superficies(ServerRequestInterface $request, ResponseInterface $response,
                                Repository\Venta $ventaRepository, Repository\Venta\Unidad $unidadRepository,
                                Service\Redis $redisService, Service\Format $formatService,
                                int $proyecto_id): ResponseInterface
    {
        $output = [
            'proyecto_id' => $proyecto_id,
            'superficies' => [
                'vendible' => 0,
                'vendido' => 0,
                'por_vender' => 0
            ],
            'formatted' => [
                'vendible' => '0m²',
                'vendido' => '0m²',
                'por_vender' => '0m²'
            ]
        ];
        $redisKey = "superficices:proyecto:{$proyecto_id}";
        try {
            $output = $this->fetchRedis($redisService, $redisKey);
        } catch (EmptyRedis) {
            try {
                $ventas = $ventaRepository->fetchActivaByProyecto($proyecto_id);
                $unidades = $unidadRepository->fetchDisponiblesByProyecto($proyecto_id);

                $output['superficies']['vendido'] = array_reduce($ventas, function($sum, Model\Venta $venta) {
                    return $sum + array_reduce($venta->propiedad()->unidades, function($sum, Model\Venta\Unidad $unidad) {
                        return $sum + $unidad->proyectoTipoUnidad->vendible();
                    });
                }, 0);
                $output['formatted']['vendido'] = "{$formatService->number($output['superficies']['vendido'], 2)}m²";
                $output['superficies']['por_vender'] = array_reduce($unidades, function($sum, Model\Venta\Unidad $unidad) {
                    return $sum + $unidad->proyectoTipoUnidad->vendible();
                });
                $output['formatted']['por_vender'] = "{$formatService->number($output['superficies']['por_vender'], 2)}m²";
                $output['superficies']['vendible'] = $output['superficies']['vendido'] + $output['superficies']['por_vender'];
                $output['formatted']['vendible'] = "{$formatService->number($output['superficies']['vendible'], 2)}m²";
                $this->saveRedis($redisService, $redisKey, $output, 60 * 60);
            } catch (EmptyResult) {}
        }
        return $this->withJson($response, $output);
    }
    public function terreno(ServerRequestInterface $request, ResponseInterface $response, Repository\Proyecto $proyectoRepository, int $proyecto_id): ResponseInterface
    {
        $body = $request->getParsedBody();
        $output = [
            'input' => $body,
            'proyecto_id' => $proyecto_id,
            'edited' => false
        ];
        try {
            $proyecto = $proyectoRepository->fetchById($proyecto_id);
            $proyectoRepository->editTerreno($proyecto, $body);
            $output['edited'] = true;
        } catch (EmptyResult) {}
        return $this->withJson($response, $output);
    }
    public function tiposUnidades(ServerRequestInterface $request, ResponseInterface $response,
                                  Repository\Proyecto $proyectoRepository,
                                  Repository\Proyecto\TipoUnidad $tipoUnidadRepository,
                                  int $proyecto_id): ResponseInterface
    {
        $output = [
            'proyecto_id' => $proyecto_id,
            'tipos' => []
        ];
        try {
            $proyecto = $proyectoRepository->fetchById($proyecto_id);
            $output['tipos'] = $tipoUnidadRepository->fetchByProyecto($proyecto->id);
        } catch (EmptyResult) {}
        return $this->withJson($response, $output);
    }
    public function brokers(ServerRequestInterface $request, ResponseInterface $response,
                            Repository\Proyecto $proyectoRepository, Service\Proyecto\Broker\Contract $contractService,
                            int $proyecto_id): ResponseInterface
    {
        $output = [
            'proyecto_id' => $proyecto_id,
            'contracts' => []
        ];
        try {
            $proyecto = $proyectoRepository->fetchById($proyecto_id);
            $output['contracts'] = $contractService->getByProject($proyecto->id);
        } catch (EmptyResult | Read $exception) {
            return $this->withError($response, $exception);
        }
        return $this->withJson($response, $output);
    }

}
